# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

"""
Numpy server for vibration features

"""


__author__ = "Deepanshu Yadav"
__copyright__ = "Copyright (c) 2021 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import logging
from collections import defaultdict, Counter
import datetime
import uuid
import json

import pandas as pd
from scipy.stats import kurtosis
from scipy.stats import skew
import scipy
import numpy as np

from foglamp.common import logger, iprpc


_LOGGER = logger.setup(__name__, level=logging.INFO)
ftr_cols = ['median', 'mean', 'std_dev', 'variance', 'RMS', 'kurtosis']


class NPServer(iprpc.InterProcessRPC):
    """ Class for offloading numpy/scipy operations from filter plugin to a separate process
    """
    def __init__(self, service_name):
        """ Initialize numpy/scipy offloading server
        Args:
            service_name: name of the server
        Returns:
        Raises:
        """
        super().__init__()  # initialize rpc server

    def compute_features(self, handle, dw):
        """ Compute features

        Args:
            A reading data
        """

        # _LOGGER.debug("dw={}".format(dw))

        t1 = datetime.datetime.now()
        df = pd.DataFrame(pd.json_normalize(dw))
        t2 = datetime.datetime.now()
        delta = t2 - t1
        _LOGGER.debug("JSON parsing & dataframe creation took {} msec".format(int(delta.total_seconds() * 1000)))

        _LOGGER.debug("Data window has {} rows, TS range = {} .. {}".format(df.shape[0], df['user_ts'][0],
                                                                            df['user_ts'].iloc[-1]))

        feature_reading = {'asset': handle['assetName'], 'timestamp': df['user_ts'].iloc[-1],
                           'key': str(uuid.uuid4()), 'readings': {}}

        t1 = datetime.datetime.now()
        for col in df.columns:
            if col.startswith('readings.ch1'):
                series = pd.Series(df[col])
                ftrs = get_features(series)
                del series
                key = col.replace("readings.", "")
                for k, v in zip(ftr_cols, ftrs):
                    feature_reading['readings'][key + '_' + k] = v

        t2 = datetime.datetime.now()
        delta = t2 - t1
        _LOGGER.debug("Features calculated in {} msec".format(int(delta.total_seconds() * 1000)))
        _LOGGER.debug("feature_reading={}".format(json.dumps(feature_reading)))

        del df

        return feature_reading

    def plugin_shutdown(self):
        """ Shutdown numpy/scipy offloading server
        Args:
        Returns:
        Raises:
            SystemExit
        """
        _LOGGER.debug("IPC/RPC server shut down with exit code 0.")
        raise SystemExit


def _kurtosis(x):
    return kurtosis(x)


def CPT5(x):
    den = len(x) * np.exp(np.std(x))
    return sum(np.exp(x)) / den


def SSC(x):
    x = np.array(x)
    x = np.append(x[-1], x)
    x = np.append(x, x[1])
    xn = x[1:len(x) - 1]
    xn_i2 = x[2:len(x)]  # xn+1
    xn_i1 = x[0:len(x) - 2]  # xn-1
    ans = np.heaviside((xn - xn_i1) * (xn - xn_i2), 0)
    return sum(ans[1:])


def wave_length(x):
    x = np.array(x)
    x = np.append(x[-1], x)
    x = np.append(x, x[1])
    xn = x[1:len(x) - 1]
    xn_i2 = x[2:len(x)]  # xn+1
    return sum(abs(xn_i2 - xn))


def norm_entropy(x):
    tresh = 2
    return sum(np.power(abs(x), tresh))


def SRAV(x):
    SRA = sum(np.sqrt(abs(x)))
    return np.power(SRA / len(x), 2)


def mean_abs(x):
    return sum(abs(x)) / len(x)


def calculate_entropy(list_values):
    counter_values = Counter(list_values).most_common()
    probabilities = [elem[1] / len(list_values) for elem in counter_values]
    entropy = scipy.stats.entropy(probabilities)
    return entropy


def calculate_statistics(list_values):
    n5 = 0.0 # np.nanpercentile(list_values, 5)
    n25 = 0.0 # np.nanpercentile(list_values, 25)
    n75 = 0.0 # np.nanpercentile(list_values, 75)
    n95 = 0.0 # np.nanpercentile(list_values, 95)
    median = np.nanpercentile(list_values, 50)
    mean = np.nanmean(list_values)
    std = np.nanstd(list_values)
    var = np.nanvar(list_values)
    rms = np.nanmean(np.sqrt(list_values ** 2))
    # New features
    kur = kurtosis(list_values)
    MeanAbs = 0.0 # mean_abs(list_values)
    norm_ent = 0.0 # norm_entropy(list_values)
    skewness = 0.0 # skew(list_values)
    CPT_5 = 0.0 # CPT5(list_values)
    SSC_1 = 0.0 # SSC(list_values)
    WL = 0.0 # wave_length(list_values)
    SRAV_1 = 0.0 # SRAV(list_values)
    return [n5, n25, n75, n95, median, mean, std, var, rms, kur, MeanAbs, norm_ent, skewness, CPT_5, SSC_1, WL, SRAV_1]


def calculate_crossings(list_values):
    zero_crossing_indices = np.nonzero(np.diff(np.array(list_values) > 0))[0]
    no_zero_crossings = len(zero_crossing_indices)
    mean_crossing_indices = np.nonzero(np.diff(np.array(list_values) > np.nanmean(list_values)))[0]
    no_mean_crossings = len(mean_crossing_indices)
    return [no_zero_crossings, no_mean_crossings]


def get_features(list_values):
    statistics = calculate_statistics(list_values)
    features = statistics[4:10] + statistics[12:13]
    return features


if __name__ == "__main__":
    _LOGGER.info("SERVING... vibration features calculation")
    rpc = NPServer('vibration feature service')
    rpc.serve()
