# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for Vibration features filter plugin

This filter collects readings for configured observation interval, then calculates statistics on these readings
and puts these statistics into a new reading.
"""

import time
import copy
import logging
import datetime
import os

from foglamp.common import logger, iprpc
import filter_ingest

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_LOGGER = logger.setup(__name__, level=logging.INFO)

# Filter specific objects
the_callback = None
the_ingest_ref = None
interval = None
shutdown_in_progress = False
data_window = []
window_start_ts = None

num_windows = 0
total_time = 0

the_rpc = None
_module_dir = os.path.dirname(__file__)

# FOGL-3453: Mean, Median, Mode, Standard Deviation, Min, Max, Variance, Range

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Vibration features filter plugin',
        'type': 'string',
        'default': 'vibration_features',
        'readonly': 'true'
    },
    'enable': {
        'description': 'Enable filter plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "3"
    },
    'interval': {
        'description': 'Observation time interval over which to calculate features',
        'type': 'float',
        'default': '10',
        'displayName': 'Observation interval (ms)',
        'order': "2"
    },
    'assetName': {
        'description': 'Asset name for reading with calculated features',
        'type': 'string',
        'default': 'VibrationFeature',
        'displayName': 'Asset name',
        'order': "1"
    }
}


def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        'name': 'vibration_features',
        'version': '1.9.2',
        'mode': "none",
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config: JSON configuration document for the Filter plugin configuration category
        ingest_ref:
        callback:
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    data = copy.deepcopy(config)

    global the_callback, the_ingest_ref, interval, the_rpc

    the_callback = callback
    the_ingest_ref = ingest_ref
    data['interval'] = float(config['interval']['value'])
    interval = data['interval']
    data['assetName'] = config['assetName']['value']

    _LOGGER.debug("plugin_init for filter vibration_features called")

    the_rpc = iprpc.IPCModuleClient("feature_server", _module_dir)

    return data


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    global interval

    interval = float(new_config['interval']['value'])
    assetName = new_config['assetName']['value']
    _LOGGER.debug("Old config for vibration_feature plugin {} \n new config {}".format(handle, new_config))
    new_handle = copy.deepcopy(new_config)
    new_handle['assetName'] = assetName

    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    global shutdown_in_progress, the_callback, the_ingest_ref, interval, data_window, the_rpc
    shutdown_in_progress = True
    time.sleep(1)
    the_callback = None
    the_ingest_ref = None
    interval = None
    data_window = []

    if the_rpc is not None:
        try:
            the_rpc.plugin_shutdown()
        except EOFError:
            _LOGGER.info("IPC/RPC server invoked for vibration features plugin shut down with exit code 0. ")
        the_rpc = None

    _LOGGER.info('Filter vibration_feature plugin shutdown.')


def plugin_ingest(handle, data):
    """ Modify readings data and pass it onward

    Args:
        handle: handle returned by the plugin initialisation call
        data: readings data
    """
    global shutdown_in_progress, the_callback, the_ingest_ref, data_window, window_start_ts, interval
    if shutdown_in_progress:
        return

    if handle['enable']['value'] == 'false':
        # Filter not enabled, just pass data onwards
        filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, data)
        return

    for elem in data:
        user_ts = elem['user_ts']
        if user_ts[-3] == ':':
            user_ts = user_ts[:-3] + user_ts[-2:]

        user_ts = datetime.datetime.strptime(user_ts, '%Y-%m-%d %H:%M:%S.%f%z').timestamp()

        if window_start_ts == None:
            window_start_ts = user_ts

        if user_ts > window_start_ts + interval/1000.0:
            t1 = datetime.datetime.now()

            if the_rpc is not None:
                feature_reading = the_rpc.compute_features(handle, data_window)
            else:
                return

            t2 = datetime.datetime.now()

            data_window.append(feature_reading)

            _LOGGER.debug("Ingesting {} readings to south service".format(len(data_window)))

            # Pass data onwards
            filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, data_window)

            t3 = datetime.datetime.now()

            del data_window, feature_reading
            data_window = []
            window_start_ts = user_ts

            delta1 = t2 - t1
            delta2 = t3 - t2
            delta = t3 - t1
            _LOGGER.debug("Data window processed in {} "
                          "msec: feature calculation took {} msec,"
                          " ingest took {} msec".format(
                                                        int(delta.total_seconds() * 1000),
                                                        int(delta1.total_seconds() * 1000),
                                                        int(delta2.total_seconds() * 1000)))

            global num_windows, total_time
            num_windows = num_windows + 1
            total_time = total_time + int(delta.total_seconds() * 1000)

            if num_windows == 1000:
                _LOGGER.debug("Last 1000 data windows processed at an avg of {} msec".format(total_time//num_windows))
                num_windows = 0
                total_time = 0

        data_window.append(elem)
